import pytest

from hypothesis import given, settings, HealthCheck  #, assume
from hypothesis import reproduce_failure  # pylint: disable=unused-import

import tempfile
import subprocess  # nosec
from io import StringIO

import pandas as pd

from tests.helpers import assert_df_equal
from tests.hypothesis_helper import dfs_min, df_data, selector

import pyranges as pr

from os import environ

if environ.get("TRAVIS"):
    max_examples = 100
    deadline = None
else:
    max_examples = 1000
    deadline = None

merge_command = "bedtools merge -o first -c 6 {} -i <(sort -k1,1 -k2,2n {})"


@pytest.mark.bedtools
@pytest.mark.parametrize("strand", [True, False])
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=dfs_min())  # pylint: disable=no-value-for-parameter
def test_merge(gr, strand):

    bedtools_strand = {True: "-s", False: ""}[strand]

    print(gr)

    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = "{}/f1.bed".format(temp_dir)
        gr.df.to_csv(f1, sep="\t", header=False, index=False)

        cmd = merge_command.format(bedtools_strand, f1)
        print(cmd)

        # ignoring bandit security warning. All strings created by test suite
        result = subprocess.check_output(  # nosec
            cmd, shell=True, executable="/bin/bash").decode()  # nosec

        if not strand:
            bedtools_df = pd.read_csv(
                StringIO(result),
                sep="\t",
                header=None,
                squeeze=True,
                usecols=[0, 1, 2],
                names="Chromosome Start End".split(),
                dtype={"Chromosome": "category"})
        else:
            bedtools_df = pd.read_csv(
                StringIO(result),
                sep="\t",
                header=None,
                squeeze=True,
                names="Chromosome Start End Strand".split(),
                dtype={"Chromosome": "category"})

    print("bedtools_df\n", bedtools_df)
    result = gr.merge(strand=strand)
    print("result\n", result.df)

    if not bedtools_df.empty:
        # need to sort because bedtools sometimes gives the result in non-natsorted chromosome order!
        if result.stranded:
            assert_df_equal(
                result.df.sort_values("Chromosome Start Strand".split()),
                bedtools_df.sort_values("Chromosome Start Strand".split()))
        else:
            assert_df_equal(
                result.df.sort_values("Chromosome Start".split()),
                bedtools_df.sort_values("Chromosome Start".split()))
    else:
        assert bedtools_df.empty == result.df.empty


@pytest.mark.parametrize("strand", [True, False])
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=df_data())  # pylint: disable=no-value-for-parameter
def test_init(gr, strand):

    c, s, e, strands = gr

    if strand:
        pr.PyRanges(chromosomes=c, starts=s, ends=e, strands=strands)
    else:
        pr.PyRanges(chromosomes=c, starts=s, ends=e)


chipseq = pr.data.chipseq()


@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(selector=selector())  # pylint: disable=no-value-for-parameter
def test_getitem(selector):

    # have these weird returns to avoid being flagged as unused code
    if len(selector) == 3:
        a, b, c = selector
        return chipseq[a, b, c]
    elif len(selector) == 2:
        a, b = selector
        return chipseq[a, b]
    elif len(selector) == 1:
        a = selector[0]
        return chipseq[a]
    elif len(selector) == 0:
        pass
    else:
        raise Exception("Should never happen")


@pytest.mark.bedtools
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=dfs_min())  # pylint: disable=no-value-for-parameter
def test_summary(gr):

    # merely testing that it does not error
    # contents are just (pandas) dataframe.describe()
    gr.summary()



import pkg_resources

import pyranges as pr


def get_example_path(basename):

    full_path = pkg_resources.resource_filename(
        "pyranges", "example_data/{}".format(basename))

    if full_path.endswith(".bam"):
        _hack_to_load_idx = pkg_resources.resource_filename(
            "pyranges", "example_data/{}.bai".format(basename))

    return full_path


def f1():

    full_path = get_example_path("f1.bed")

    return pr.read_bed(full_path)


def f2():

    full_path = get_example_path("f2.bed")

    return pr.read_bed(full_path)


def chipseq():

    full_path = get_example_path("chipseq.bed")

    return pr.read_bed(full_path)


def epigenome_roadmap():

    full_path = get_example_path("epigenome_roadmap.bed")

    return pr.read_bed(full_path)


def chipseq_background():

    full_path = get_example_path("chipseq_background.bed")

    return pr.read_bed(full_path)


def aorta():

    full_path = get_example_path("aorta.bed")

    return pr.read_bed(full_path)


def aorta2():

    full_path = get_example_path("aorta2.bed")

    return pr.read_bed(full_path)


def ensembl_gtf():

    full_path = get_example_path("ensembl.gtf")

    return pr.read_gtf(full_path)


def control_bam():

    full_path = get_example_path("control.bam")

    return pr.read_bam(full_path)



# import pytest

# from hypothesis import given, settings, HealthCheck
# from hypothesis import reproduce_failure  # pylint: disable=unused-import

# from itertools import product
# import tempfile
# import subprocess  # nosec
# from io import StringIO

# import pandas as pd

# from tests.helpers import assert_df_equal
# from tests.hypothesis_helper import dfs_min2, dfs_min, dfs_no_min

# from os import environ

# from itertools import product

# if environ.get("TRAVIS"):
#     max_examples = 100
#     deadline = None
# else:
#     max_examples = 100
#     deadline = None

# strandedness = [False, "same", "opposite"]

# binary_methods = [
#     "set_union", "set_intersect", "overlap", "nearest", "intersect",
#     "subtract", "join"
# ]

# unary_methods = ["merge", "sort"]

# method_chain = product(binary_methods, binary_methods)
# # cannot start with an operation that makes pyrange unstranded and then try a stranded op
# strandedness_chain = list(product(["same", "opposite"], strandedness)) + list(
#     product(strandedness, [None]))

# @pytest.mark.bedtools
# @pytest.mark.parametrize("strandedness_chain,method_chain",
#                          product(strandedness_chain, method_chain))
# @settings(
#     max_examples=max_examples,
#     deadline=deadline,
#     suppress_health_check=HealthCheck.all())
# @given(gr=dfs_no_min(), gr2=dfs_no_min(), gr3=dfs_no_min())  # pylint: disable=no-value-for-parameter
# @reproduce_failure('4.15.0', b'AXicY2RAA4wQDEFIYsh8AAD/AAg=')
# def test_three_in_a_row(gr, gr2, gr3, strandedness_chain, method_chain):

#     # https://stackoverflow.com/questions/26778079/valueerror-ndarray-is-not-c-contiguous-in-cython
#     # df.Start.values.flags["C_CONTIGUOUS"]

#     s1, s2 = strandedness_chain
#     f1, f2 = method_chain

#     print(s1, s2)
#     print(f1, f2)

#     m1 = getattr(gr, f1)
#     gr2 = m1(gr2, strandedness=s1)
#     print("first finished!")
#     m2 = getattr(gr2, f2)
#     gr3 = m2(gr3, strandedness=s2)
